#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/wdt.h>
#include <avr/builtins.h>
#define F_CPU 8000000UL
#include <util/delay.h>
#include <stdlib.h>
#include <string.h>

// Macros
#define MESSAGE_LENGTH 		8
#define TRX_FRAME_BUFFER(index) (*(volatile uint8_t *)(0x180 + (index)))
#define FRAME_TIME 		0x61A8
#define ACTIVE_TIME		0x30D4
#define SLEEP_TIME		(FRAME_TIME) - (ACTIVE_TIME)

#define STATE_SEARCHING 	0
#define STATE_SMAC 		1

#define MESSAGE_NDIS	 	1
#define MESSAGE_RTS	 	2
#define MESSAGE_CTS	 	3
#define MESSAGE_PACKET	 	4

#define QSIZE			5
#define PAYLOAD_LENGTH		3


// Structure Definitions
typedef struct
{
	uint8_t level;
} ndis_t;
typedef struct
{
	uint8_t to;
} flag_t;
typedef struct
{
	uint8_t from_level;
	uint8_t data[PAYLOAD_LENGTH];
} packet_t;
typedef struct
{
	uint8_t msgType;	
	uint8_t from;
	union Payload 
	{
		uint8_t payload[MESSAGE_LENGTH - 1];
		ndis_t ndis;
		flag_t flag;
		packet_t packet;
	} payload;
	uint16_t crc;
} Frame_t;


// Variables
uint32_t rtcCompareValue = 0;
uint8_t nod1_address __attribute__((section(".data"))) = 0;
uint8_t node_address __attribute__((section(".data"))) = 0;

uint8_t state;
uint8_t receivedMsg;
uint8_t timerTick;
uint8_t active = 0;
uint8_t level;
uint8_t next_hop_to_base;
int awaiting_cts_from = -1;

uint16_t counter;
Frame_t receivedFrame;

Frame_t frameQueue[QSIZE];
uint8_t first = 0,last= 0;

// Function Declarations
void setState(uint8_t state);

// Function Definitions
void initRF(void)
{
  setState(CMD_TRX_OFF);

  IRQ_STATUS = 0xff;
  IRQ_MASK_struct.rx_end_en = 1;

  setState(RX_ON);
}

void initTimer(uint32_t initialWait)
{
	ASSR = 0;
	ASSR = _BV(AS2); // enable asynchronous mode, with external oscillator (32.768kHz in our case) 

	rtcCompareValue = (uint32_t)initialWait;
	
	SCOCR1HH = (uint8_t) ((rtcCompareValue >> 24) & 0x000000ff);
	SCOCR1HL = (uint8_t) ((rtcCompareValue >> 16) & 0x000000ff);
	SCOCR1LH = (uint8_t) ((rtcCompareValue >> 8) & 0x000000ff);
	SCOCR1LL = (uint8_t) ((rtcCompareValue) & 0x000000ff);

	SCCR0 |= _BV(SCCKSEL);

	SCCR0 |= _BV(SCRES) | _BV(SCTSE) | _BV(SCEN) | _BV(SCCKSEL); 	// enable symbol counter, with TOSC1/2 clock (32.768kHz)



	SCCNTHH = 0x00;
	SCCNTHL = 0x00;
	SCCNTLH = 0x00;
	SCCNTLL = 0x00;
	SCCR0 |= _BV(SCCKSEL);


	while (SCSR & _BV(SCBSY));
	SCIRQM |= _BV(IRQMCP1); 			// enable compare match 1 IRQ
}
void setTimer(uint32_t compare)
{
	rtcCompareValue = (uint32_t) compare;
	
	SCOCR1HH = (uint8_t) ((rtcCompareValue >> 24) & 0x000000ff);
	SCOCR1HL = (uint8_t) ((rtcCompareValue >> 16) & 0x000000ff);
	SCOCR1LH = (uint8_t) ((rtcCompareValue >> 8) & 0x000000ff);
	SCOCR1LL = (uint8_t) ((rtcCompareValue) & 0x000000ff);

	while (SCSR & _BV(SCBSY));

}
void delayTimer(uint32_t delay)
{
	rtcCompareValue += (uint32_t) delay;
	
	SCOCR1HH = (uint8_t) ((rtcCompareValue >> 24) & 0x000000ff);
	SCOCR1HL = (uint8_t) ((rtcCompareValue >> 16) & 0x000000ff);
	SCOCR1LH = (uint8_t) ((rtcCompareValue >> 8) & 0x000000ff);
	SCOCR1LL = (uint8_t) ((rtcCompareValue) & 0x000000ff);

	while (SCSR & _BV(SCBSY));
}
uint32_t readCounter()
{
	uint32_t ret;
	ret = SCCNTLL;
	ret += ((uint32_t)SCCNTLH) << 8;
	ret += ((uint32_t)SCCNTHL) << 16;
	ret += ((uint32_t)SCCNTHH) << 24;
	return ret;
}
uint32_t readTimestamp()
{
	uint32_t ret;
	ret = SCTSRLL;
	ret += ((uint32_t)SCTSRLH) << 8;
	ret += ((uint32_t)SCTSRHL) << 16;
	ret += ((uint32_t)SCTSRHH) << 24;
	return ret;
}


void setState(uint8_t state)
{

	TRX_STATE = CMD_FORCE_TRX_OFF;
	while (TRX_STATUS_struct.trx_status == STATE_TRANSITION_IN_PROGRESS);
	switch(TRX_STATUS_struct.trx_status)
	{
		case TRX_OFF: TRX_STATE = state; break;
		case BUSY_TX:
		case BUSY_RX: 
			      if (state == CMD_PLL_ON)
			      {
				      TRX_STATE = CMD_FORCE_PLL_ON; 
				      break;
			      }
			      else 
				      TRX_STATE = CMD_FORCE_TRX_OFF; 
			      while (TRX_STATUS_struct.trx_status == STATE_TRANSITION_IN_PROGRESS);
			      TRX_STATE = state;
			      break;
	}
	while (state != TRX_STATUS_struct.trx_status);
}

void timer_task()
{
	if (counter==1 && node_address==0)
	{
		level = 0;
		sendFirstNdisMessage();
	}
	if (active && first != last)
	{
		if (frameQueue[first].msgType!= MESSAGE_NDIS)
			sendFlagMessage(MESSAGE_RTS, next_hop_to_base);
		else
		{
			popMessage();
		}
	}


	if (counter == 20 && node_address==5)
	{
		uint8_t payload[3];
		payload[0] = 42;
		payload[1] = 42;
		payload[2] = 42;
		pushPacketToBase(payload);
	}
		
	
		
}
void receive_task()
{
	switch(state)
	{
		case STATE_SEARCHING:
			if (receivedFrame.msgType == MESSAGE_NDIS)
			{
				state = STATE_SMAC;
				level = receivedFrame.payload.ndis.level+1;
				next_hop_to_base = receivedFrame.from;
				DDRD = level;
				pushNdisMessage();
				pushNdisMessage();
			}
			break;
		case STATE_SMAC:
			if (receivedFrame.msgType == MESSAGE_RTS && receivedFrame.payload.flag.to == node_address)
				sendFlagMessage(MESSAGE_CTS,receivedFrame.from);
			else if ((receivedFrame.msgType == MESSAGE_CTS) && receivedFrame.payload.flag.to == node_address)
				popMessage();
			else if (receivedFrame.msgType == MESSAGE_PACKET && receivedFrame.payload.packet.from_level > level
					 && node_address != 0) // node 0 consumes packets - base station
				pushPacket();

			break;
	}
}
ISR(SCNT_CMP1_vect)
{
	counter++;

	active = 1 - active;

	if (!active) 
	{
		PORTE |= _BV(PE4);
		delayTimer(SLEEP_TIME);
	}
	else 
	{
		PORTE &= ~_BV(PE4);
		delayTimer(ACTIVE_TIME);
	}
	timerTick = 1;
}

ISR(TRX24_RX_END_vect)
{
	if (!(PHY_RSSI & _BV(RX_CRC_VALID)))
		return;

	memcpy(&receivedFrame,(void*)&TRXFBST, sizeof(Frame_t));
	receivedMsg = 1;
}

int main()
{
	PORTE = 0xff;
	DDRE = 0xff;

	if (node_address == 0)
		state = STATE_SMAC;
	else
		state = STATE_SEARCHING;
	srand(node_address);

	initTimer(100);
	initRF();
	sei();

	while(1) 
	{
		if (receivedMsg)
		{
			receivedMsg = 0;
			receive_task();
		}
		else if (timerTick)
		{
			timerTick = 0;
			timer_task();
		}
		asm("sleep");
	}
	while(1);
}
